/*
 * Copyright (C) 2011 Aastha Mehta
 * Copyright (C) 2011 Mike Christie
 *
 * maintained by open-iscsi@googlegroups.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 */
#ifndef __KERN_ERR_TABLE_H__
#define __KERN_ERR_TABLE_H__

extern const char *kern_err_code_to_string(int);
#endif
