#ifndef ISCSI_VERSION_DEF
#define ISCSI_VERSION_DEF

/*
 * iSCSI tools version.
 * This may not be the same value as the kernel versions because
 * some other maintainer could merge a patch without going through us
 */
#define ISCSI_VERSION_STR	"2.0-877"
#define ISCSI_VERSION_FILE	"/sys/module/scsi_transport_iscsi/version"

#endif
