/*
 * iSCSI timer
 *
 * Copyright (C) 2002 Cisco Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 */
#ifndef ISCSI_TIMER_H
#define ISCSI_TIMER_H

struct timeval;

extern void iscsi_timer_clear(struct timeval *timer);
extern void iscsi_timer_set(struct timeval *timer, int seconds);
extern int iscsi_timer_expired(struct timeval *timer);
extern int iscsi_timer_msecs_until(struct timeval *timer);

#endif
